#include <errno.h>

#include "emu32.h"
#include "emu-hcall.c"

/* --------------------------------------------------------------------- */

#define HCALL_HANDLED  0
#define HCALL_FORWARD -1

typedef int32_t (*xen_hcall)(uint32_t *args);
static int32_t multicall(uint32_t *args);

/* --------------------------------------------------------------------- */

static int32_t update_va_mapping(uint32_t *args)
{
    uint32_t va    = args[0];
    uint64_t val   = args[1] | ((uint64_t)args[2] << 32 );
    uint32_t flags = args[3];
    void *map = NULL;

    if (is_pae()) {
	uint64_t *ptr = find_pte_pae_lpt(va);
	uint64_t pte;

	if (va >= XEN_M2P_PAE)
	    goto inval;
	if (memcpy_pf(&pte, ptr, sizeof(pte)) < 0) {
	    printk("%s: lpt rd fault: va %" PRIx32 " val %" PRIx64 "\n",
		   __FUNCTION__, va, val);
	    goto inval;
	}
	if ((pte & ~_PAGE_RW) == val && !(val & _PAGE_USER)) {
	    /* ignore make_readonly */
	    goto doflags;
	}
	pte = val;
	if (0 == memcpy_pf(ptr, &pte, sizeof(pte)))
	    goto doflags;
	printk("%s: lpt wr fault: va %" PRIx32 " val %" PRIx64 "\n",
	       __FUNCTION__, va, val);
	ptr = find_pte_pae_map(va);
	if (NULL == ptr)
	    goto inval;
	map = ptr;
	if (0 == memcpy_pf(ptr, &pte, sizeof(pte)))
	    goto doflags;
	printk("%s: map wr fault: va %" PRIx32 " val %" PRIx64 "\n",
	       __FUNCTION__, va, val);
	goto inval;
    } else {
	uint32_t *ptr = find_pte_32_lpt(va);
	uint32_t pte;

	if (va >= XEN_M2P_32)
	    goto inval;
	if (memcpy_pf(&pte, ptr, sizeof(pte)) < 0)
	    goto inval;
	if ((pte & ~_PAGE_RW) == val && !(val & _PAGE_USER)) {
	    /* ignore make_readonly */
	    goto doflags;
	}
	pte = val;
	if (0 == memcpy_pf(ptr, &pte, sizeof(pte)))
	    goto doflags;
	printk("%s: lpt wr fault: va %" PRIx32 " val %" PRIx64 "\n",
	       __FUNCTION__, va, val);
	ptr = find_pte_32_map(va);
	if (NULL == ptr)
	    goto inval;
	map = ptr;
	if (0 == memcpy_pf(ptr, &pte, sizeof(pte)))
	    goto doflags;
	printk("%s: map wr fault: va %" PRIx32 " val %" PRIx64 "\n",
	       __FUNCTION__, va, val);
	goto inval;
    }

doflags:
    if (map)
	free_page(map);
    switch (flags & UVMF_FLUSHTYPE_MASK) {
    case UVMF_NONE:
	break;
    case UVMF_TLB_FLUSH:
	flush_tlb();
	break;
    case UVMF_INVLPG:
	flush_tlb_addr(va);
	break;
    }
    return 0;

inval:
    if (map)
	free_page(map);
    return -EINVAL;
}

static int32_t mmu_update(uint32_t *args)
{
    uint64_t *reqs = (void*)args[0];
    uint32_t count = args[1];
//    uint32_t *done = (void*)args[2];
    uint32_t dom   = args[3];
    int i;

    if (dom != DOMID_SELF) {
	printk("%s: foreigndom not supported\n", __FUNCTION__);
	return -ENOSYS;
    }

    for (i = 0; i < count; i++) {
	switch (reqs[0] & 3) {
	case MMU_NORMAL_PT_UPDATE:
	    if (is_pae()) {
		uint64_t *pte = map_pae(reqs[0]);
		*pte = reqs[1];
		free_page(pte);
	    } else {
		uint32_t *pte = map_32(reqs[0]);
		*pte = reqs[1];
		free_page(pte);
	    }
	    break;
	case MMU_MACHPHYS_UPDATE:
	    printk("%s: machphys: %" PRIx64 " %" PRIx64 "\n",
		   __FUNCTION__, reqs[0], reqs[1]);
	    panic("FIXME: MMU_MACHPHYS_UPDATE not implemented yet");
	    break;
	default:
	    return -ENOSYS;
	}
	reqs += 2;
    }
    return 0;
}

static int32_t mmuext_op(uint32_t *args)
{
    struct mmuext_op *uops = (void*)args[0];
    uint32_t count = args[1];
//    uint32_t *done = (void*)args[2];
    uint32_t dom   = args[3];
    int i;

    if (dom != DOMID_SELF) {
	printk("%s: foreigndom not supported\n", __FUNCTION__);
	return -ENOSYS;
    }

    for (i = 0; i < count; i++, uops++) {
	switch (uops->cmd) {
	case MMUEXT_PIN_L1_TABLE:
	case MMUEXT_PIN_L2_TABLE:
	case MMUEXT_PIN_L3_TABLE:
	case MMUEXT_PIN_L4_TABLE:
	case MMUEXT_UNPIN_TABLE:
	    /* ignore */
	    break;
	case MMUEXT_SET_LDT:
	    printk("%s: FIXME: ignore SET_LDT (va %lx, nr %d)\n", __FUNCTION__,
		   uops->arg1.linear_addr, uops->arg2.nr_ents);
	    if (uops->arg2.nr_ents)
		panic("ldt gets used");
	    break;
	case MMUEXT_INVLPG_LOCAL:
	case MMUEXT_INVLPG_MULTI:     /* no SMP yet */
	    flush_tlb_addr(uops->arg1.linear_addr);
	    break;
	case MMUEXT_TLB_FLUSH_LOCAL:
	case MMUEXT_TLB_FLUSH_ALL:    /* no SMP yet */
	case MMUEXT_TLB_FLUSH_MULTI:  /* no SMP yet */
	    flush_tlb();
	    break;
	case MMUEXT_NEW_BASEPTR:
	    update_emu_mappings(uops->arg1.mfn);
	    write_cr3(frame_to_addr(uops->arg1.mfn));
	    break;
	default:
	    printk("%s: FIXME: unknown %d\n", __FUNCTION__, uops->cmd);
	    return -ENOSYS;
	}
    }
    return 0;
}

/* --------------------------------------------------------------------- */

static xen_hcall hcalls[XEN_HCALL_MAX] = {
    [ __HYPERVISOR_update_va_mapping ] = update_va_mapping,
    [ __HYPERVISOR_mmu_update ]        = mmu_update,
    [ __HYPERVISOR_mmuext_op ]         = mmuext_op,
    [ __HYPERVISOR_update_descriptor ] = update_descriptor,
    [ __HYPERVISOR_stack_switch ]      = stack_switch,
    [ __HYPERVISOR_multicall ]         = multicall,
    [ __HYPERVISOR_fpu_taskswitch ]    = fpu_taskswitch,
    [ __HYPERVISOR_grant_table_op ]    = grant_table_op,
    [ __HYPERVISOR_xen_version ]       = xen_version,
    [ __HYPERVISOR_vm_assist ]         = vm_assist,

    [ __HYPERVISOR_physdev_op ]        = error_noperm,
    [ __HYPERVISOR_physdev_op_compat ] = error_noperm,
    [ __HYPERVISOR_set_debugreg ]      = error_noperm,

    [ __HYPERVISOR_nmi_op ]            = error_nosys,
};

static int32_t multicall(uint32_t *args)
{
    struct multicall_entry *calls = (void*)args[0];
    uint32_t i, count = args[1];
    uint32_t margs[6];

    for (i = 0; i < count; i++) {
	if (!hcalls[calls[i].op])
	    panic("unknown hypercall in multicall list");
	emu_mcalls[calls[i].op]++;
	margs[0] = calls[i].args[0];
	margs[1] = calls[i].args[1];
	margs[2] = calls[i].args[2];
	margs[3] = calls[i].args[3];
	margs[4] = calls[i].args[4];
	margs[5] = calls[i].args[5];
	calls[i].result = hcalls[calls[i].op](margs);
    }
    return 0;
}

asmlinkage void do_hypercall(struct regs_32 regs)
{
    uint32_t args[6];
    uint32_t retval = -ENOSYS;

    if (regs.eax >= XEN_HCALL_MAX) {
	/* invalid hypercall number */
	goto handled;
    }
    if (!hcalls[regs.eax]) {
	/* no hypercall handler */
	goto forward;
    }

    /* do call */
    emu_hcalls[regs.eax]++;
    args[0] = regs.ebx;
    args[1] = regs.ecx;
    args[2] = regs.edx;
    args[3] = regs.esi;
    args[4] = regs.edi;
    args[5] = regs.ebp;
    retval = hcalls[regs.eax](args);
    if (-ENOSYS == retval)
	goto forward;

handled:
    regs.error = HCALL_HANDLED;
    regs.eax = retval;
    return;

forward:
    regs.error = HCALL_FORWARD;
    return;
}
